"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _core = require("@babel/core");

var _numberToIdentifer = _interopRequireDefault(require("./numberToIdentifer"));

var _nativeElementAttributes = _interopRequireDefault(require("./nativeElementAttributes"));

var _attributesByElement = _interopRequireDefault(require("./attributesByElement"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

const parse = ([typeNode, propNode, ...children], helpers) => {
  const props = new Map();
  let dynamicProps = undefined;
  let keyNode = undefined;

  if (_core.types.isObjectExpression(propNode)) {
    for (const prop of propNode.properties) {
      if (_core.types.isIdentifier(prop.key)) {
        if (prop.key.name === "key") {
          keyNode = prop.value;
        } else {
          props.set(prop.key.name, prop.value);
        }
      }
    }
  } else if (!_core.types.isNullLiteral(propNode)) {
    dynamicProps = propNode;
  }

  const isFragment = helpers.getType(typeNode) === "Fragment";
  return {
    kind: isFragment ? "Fragment" : _core.types.isStringLiteral(typeNode) ? "html" : "Component",
    typeNode,
    props,
    keyNode,
    dynamicProps,
    children
  };
};

class ContextScope {
  constructor(helpers) {
    this.contextName = helpers.getUniqueIdentifier("context");
    this._slotIndex = 0;
    this._captured = [];
    this.helpers = helpers;
  }

  context() {
    return this.contextName;
  }

  type() {
    return _core.types.memberExpression(this.contextName, _core.types.identifier("_"));
  }

  unmount() {
    return _core.types.memberExpression(this.contextName, _core.types.identifier("$"));
  }

  root() {
    return _core.types.memberExpression(this.contextName, _core.types.identifier("$$"));
  }

  _slotIndexToName(i, name) {
    return this.helpers.debug ? `${(0, _numberToIdentifer.default)(i)}_${name}` : (0, _numberToIdentifer.default)(i);
  }

  createSlot(name) {
    const i = this._slotIndex++;
    return _core.types.memberExpression(this.contextName, _core.types.identifier(this._slotIndexToName(i, name)));
  }

  createSlotName(name) {
    const i = this._slotIndex++;
    return _core.types.stringLiteral(this._slotIndexToName(i, name));
  }

  createCapturedLocal(name, capturedExpression) {
    const ident = this.helpers.getUniqueIdentifier(name);

    this._captured.push(_core.types.variableDeclaration("const", [_core.types.variableDeclarator(ident, capturedExpression)]));

    return ident;
  }

  getCaptured() {
    return this._captured;
  }

  importHelper(exportName) {
    return this.helpers.importHelper(exportName);
  }

  createGlobalToken(name) {
    const ident = this.helpers.getUniqueIdentifier(name);
    this.helpers.addGlobal(_core.types.variableDeclaration("const", [_core.types.variableDeclarator(ident, _core.types.objectExpression([]))]));
    return ident;
  }

}

const createNativeElement = (element, scope) => {
  const node = scope.createSlot("node");
  const createStatements = [];
  const updateStatements = [];
  const unmountStatements = [];
  let key = undefined;
  let children = element.children;

  if (element.kind === "Fragment") {// do not create a node, a fragment will be created in the children handler
  } else if (element.kind === "html") {
    const createElement = scope.importHelper("createElement");
    createStatements.push(_core.types.expressionStatement(_core.types.assignmentExpression("=", node, _core.types.callExpression(createElement, [element.typeNode]))));
    const elementName = element.typeNode.value;
    let props = element.props;

    if (element.dynamicProps) {
      const propsLocal = scope.createCapturedLocal("props", element.dynamicProps);
      props = new Map();
      const attributes = _attributesByElement.default[element.typeNode.value];

      for (const key of attributes) {
        if (key === "children") {
          if (children.length === 0) {
            children = [_core.types.memberExpression(propsLocal, _core.types.identifier(key))];
          }
        } else {
          props.set(key, _core.types.memberExpression(propsLocal, _core.types.identifier(key)));
        }
      }
    }

    for (const [key, value] of props) {
      (0, _nativeElementAttributes.default)(scope, node, elementName, key, value, element.props, createStatements, updateStatements, unmountStatements);
    }
  } else {
    console.log(element);
    throw new Error(`Unsupported element kind ${element.kind}`);
  }

  if (children.length > 0 || element.kind === "Fragment") {
    const childrenCreate = [];

    for (const child of children) {
      if (_core.types.isStringLiteral(child)) {
        childrenCreate.push(child);
        continue;
      }

      if (_core.types.isCallExpression(child) && scope.helpers.getType(child.callee) === "createElement") {
        // Inline element
        const parsedChild = parse(child.arguments, scope.helpers);

        if (parsedChild.kind !== "Component") {
          // inline native element
          const childElement = createNativeElement(parsedChild, scope);

          for (const statement of childElement.create) {
            createStatements.push(statement);
          }

          for (const statement of childElement.update) {
            updateStatements.push(statement);
          }

          for (const statement of childElement.unmount) {
            unmountStatements.push(statement);
          }

          childrenCreate.push(childElement.node);
          continue;
        }
      }

      const childLocal = scope.createCapturedLocal("child", child);
      const childNode = scope.createSlotName(`child_node`);
      createStatements.push(_core.types.expressionStatement(_core.types.callExpression(scope.importHelper("renderInternal"), [scope.context(), childLocal, childNode, _core.types.numericLiteral(1)])));
      updateStatements.push(_core.types.expressionStatement(_core.types.callExpression(scope.importHelper("renderInternal"), [scope.context(), childLocal, childNode, _core.types.numericLiteral(0)])));

      if (element.kind === "Fragment") {
        updateStatements.push(_core.types.expressionStatement(_core.types.assignmentExpression("=", _core.types.memberExpression(node, _core.types.numericLiteral(childrenCreate.length), true), _core.types.memberExpression(scope.context(), childNode, true))));
      }

      unmountStatements.push(_core.types.expressionStatement(_core.types.callExpression(scope.importHelper("unmountInternal"), [_core.types.memberExpression(scope.context(), _core.types.identifier(childNode.value + "_"))])));
      childrenCreate.push(_core.types.memberExpression(scope.context(), childNode, true));
    }

    if (element.kind === "Fragment") {
      createStatements.push(_core.types.expressionStatement(_core.types.assignmentExpression("=", node, _core.types.callExpression(scope.importHelper("createFragment"), [_core.types.arrayExpression(childrenCreate)]))));
    } else {
      createStatements.push(_core.types.expressionStatement(_core.types.callExpression(scope.importHelper("renderChildren"), [node, _core.types.arrayExpression(childrenCreate)])));
    }
  }

  return {
    create: createStatements,
    update: updateStatements,
    unmount: unmountStatements,
    node,
    key
  };
};

const createComponentInstructions = (element, helpers) => {
  const props = [];

  for (const [key, value] of element.props) {
    props.push(_core.types.objectProperty(_core.types.identifier(key), value));
  }

  if (element.children.length > 1) {
    const fragmentElement = {
      kind: "Fragment",
      children: element.children
    };
    props.push(_core.types.objectProperty(_core.types.identifier("children"), createNativeInstructions(fragmentElement, helpers)));
  } else if (element.children.length === 1) {
    props.push(_core.types.objectProperty(_core.types.identifier("children"), element.children[0]));
  }

  let instructions = _core.types.callExpression(helpers.importHelper("hooks"), [element.typeNode, _core.types.objectExpression(props)]);

  instructions = ensureKey(instructions, helpers, element.keyNode);
  return instructions;
};

const ensureCapture = (expr, scope) => {
  const caputured = scope.getCaptured();
  if (caputured.length === 0) return expr;
  return _core.types.callExpression(_core.types.arrowFunctionExpression([], _core.types.blockStatement(caputured.concat(_core.types.returnStatement(expr)))), []);
};

const ensureKey = (expr, helpers, key) => {
  if (!key) return expr;
  const withKey = helpers.importHelper("withKey");
  return _core.types.callExpression(withKey, [key, expr]);
};

const createNativeInstructions = (parsedElement, helpers) => {
  const scope = new ContextScope(helpers);
  const element = createNativeElement(parsedElement, scope);
  const token = scope.createGlobalToken("instructions");

  const test = _core.types.binaryExpression("!==", scope.type(), token);

  let instructions = _core.types.arrowFunctionExpression([scope.context()], _core.types.blockStatement([_core.types.ifStatement(test, _core.types.blockStatement([_core.types.ifStatement(scope.unmount(), _core.types.expressionStatement(_core.types.callExpression(scope.unmount(), []))), _core.types.expressionStatement(_core.types.assignmentExpression("=", scope.unmount(), element.unmount.length > 0 ? _core.types.arrowFunctionExpression([], _core.types.blockStatement(element.unmount)) : _core.types.nullLiteral())), _core.types.expressionStatement(_core.types.assignmentExpression("=", scope.type(), token))].concat(element.create)), _core.types.blockStatement(element.update)), _core.types.returnStatement(element.node)]));

  if (parsedElement.keyNode) {
    instructions = ensureKey(instructions, helpers, parsedElement.keyNode);
  }

  instructions = ensureCapture(instructions, scope);
  return instructions;
};

var _default = (node, helpers) => {
  const parsedElement = parse(node.arguments, helpers);

  if (parsedElement.kind === "Component") {
    return createComponentInstructions(parsedElement, helpers);
  } else {
    return createNativeInstructions(parsedElement, helpers);
  }
};

exports.default = _default;